/*
 * Written by Doug Lea with assistance from members of JCP JSR-166
 * Expert Group and released to the public domain, as explained at
 * http://creativecommons.org/licenses/publicdomain
 */
/*
 * Modifications by Neal Gafter illustrating the use of closures,
 * and released into the public domain, as explained at
 * http://creativecommons.org/licenses/publicdomain.
 */

package jsr166z.forkjoin;
import java.util.*;

/**
 * Interfaces and utilities describing per-element operations used
 * within parallel methods on aggregates. This class provides type
 * names for common operation signatures accepting zero, one or two
 * arguments, and returning zero or one results, for parameterized
 * types, as well as specializations to <tt>int</tt>, <tt>long</tt>,
 * and <tt>double</tt>. (Lesser used types like <tt>short</tt> are
 * absent.)
 * 
 * <p>In addition to stated signatures, implementations of these
 * interfaces must work safely in parallel. In general, this means
 * methods should operate only on their arguments, and should not rely
 * on ThreadLocals, unsafely published globals, or other unsafe
 * constructions. Additionally, they should not block waiting for
 * synchronization.
 *
 * <p> This class also contains a few commonly used implementations
 * of some of these interfaces
 *
 * <p>This class is normally best used via <tt>import static</tt>.
 */
public class Ops {

    /**
     * A reducer that adds two double elements
     */
    public static final { double, double => double } doubleAdder = { double x, double y => x + y };

    /**
     * A reducer that adds two double elements
     */
    public static final { long, long => long } longAdder = { long x, long y => x + y };
 
    /**
     * A reducer that adds two int elements
     */
    public static final { int, int => int } intAdder = { int x, int y => x + y };


    // comparators

    /**
     * A Comparator for Comparable.objects
     */
    static final class NaturalComparator<T extends Comparable<? super T>> 
        implements Comparator<T> {
        /**
         * Creates a NaturalComparator for the given element type
         * @param type the type
         */
        NaturalComparator(Class<T> type) {}

        public int compare(T a, T b) {
            return a.compareTo(b);
        }
    }

    /**
     * A reducer returning the maximum of two Comparable elements,
     * treating null as less than any non-null element.
     */
    public static final class 
        NaturalMaxReducer<T extends Comparable<? super T>>
        implements { T, T => T } {
        /**
         * Creates a NaturalMaxReducer for the given element type
         * @param type the type
         */
        NaturalMaxReducer(Class<T> type) {}
        public T invoke(T a, T b) {
            return (a != null &&
                    (b == null || a.compareTo(b) >= 0))? a : b;
        }
    }

    /**
     * A reducer returning the minimum of two Comparable elements,
     * treating null as less than any non-null element.
     */
    public static final class 
        NaturalMinReducer<T extends Comparable<? super T>>
        implements { T, T => T } {
        /**
         * Creates a NaturalMinReducer for the given element type
         * @param type the type
         */
        NaturalMinReducer(Class<T> type) {}
        public T invoke(T a, T b) {
            return (a != null &&
                    (b == null || a.compareTo(b) <= 0))? a : b;
        }
    }
    
    /**
     * A reducer returning the maximum of two elements, using the
     * given comparator, and treating null as less than any non-null
     * element.
     */
    public static final class MaxReducer<T> implements { T, T => T } {
        private final Comparator<? super T> comparator;
        public MaxReducer(Comparator<? super T> comparator) {
            this.comparator = comparator;
        }
        public T invoke(T a, T b) {
            return (a != null &&
                    (b == null || comparator.compare(a, b) >= 0))? a : b;
        }
    }
    
    /**
     * A reducer returning the minimum of two elements, using the
     * given comparator, and treating null as greater than any non-null
     * element.
     */
    public static final class MinReducer<T> implements { T, T => T } {
        private final Comparator<? super T> comparator;
        public MinReducer(Comparator<? super T> comparator) {
            this.comparator = comparator;
        }
        public T invoke(T a, T b) {
            return (a != null &&
                    (b == null || comparator.compare(a, b) <= 0))? a : b;
        }
    }
    
    /**
     * A comparator for doubles relying on natural ordering
     */
    public static final { double, double => int } naturalDoubleComparator = { double a, double b => Double.compare(a, b) };

    /**
     * A reducer returning the maximum of two double elements, using
     * natural comparator
     */
    public static final { double, double => double } naturalDoubleMaxReducer = { double a, double b => a >= b ? a : b };

    /**
     * A reducer returning the minimum of two double elements,
     * using natural comparator
     */
    public static final { double, double => double } naturalDoubleMinReducer = { double a, double b => a <= b? a : b };
    
    /**
     * A reducer returning the maximum of two double elements,
     * using the given comparator
     */
    public static final class DoubleMaxReducer implements { double, double => double } {
        final { double, double => int } comparator;
        /**
         * Creates a DoubleMaxReducer using the given comparator
         */
        public DoubleMaxReducer({ double, double => int } comparator) {
            this.comparator = comparator;
        }
        public double invoke(double a, double b) { 
            return (comparator.invoke(a, b) >= 0) ? a : b; 
        }
    }

    /**
     * A reducer returning the minimum of two double elements,
     * using the given comparator
     */
    public static final class DoubleMinReducer implements { double, double => double } {
        final { double, double => int } comparator;
        /**
         * Creates a DoubleMinReducer using the given comparator
         */
        public DoubleMinReducer({ double, double => int } comparator) {
            this.comparator = comparator;
        }
        public double invoke(double a, double b) { 
            return (comparator.invoke(a, b) <= 0)? a : b; 
        }
    }

    /**
     * A comparator for longs relying on natural ordering
     */
    public static final { long, long => int } naturalLongComparator = { long a, long b => (a < b) ? -1 : ((a > b)? 1 : 0) };

    /**
     * A reducer returning the maximum of two long elements, using
     * natural comparator
     */
    public static final { long, long => long } naturalLongMaxReducer = { long a, long b => a >= b? a : b };

    /**
     * A reducer returning the minimum of two long elements,
     * using natural comparator
     */
    public static final { long, long => long } naturalLongMinReducer = { long a, long b => a <= b? a : b };

    /**
     * A reducer returning the maximum of two long elements,
     * using the given comparator
     */
    public static final class LongMaxReducer implements { long, long => long } {
        final { long, long => int } comparator;
        /**
         * Creates a LongMaxReducer using the given comparator
         */
        public LongMaxReducer({ long, long => int } comparator) {
            this.comparator = comparator;
        }
        public long invoke(long a, long b) { 
            return (comparator.invoke(a, b) >= 0) ? a : b; 
        }
    }

    /**
     * A reducer returning the minimum of two long elements,
     * using the given comparator
     */
    public static final class LongMinReducer implements { long, long => long } {
        final { long, long => int } comparator;
        /**
         * Creates a LongMinReducer using the given comparator
         */
        public LongMinReducer({ long, long => int } comparator) {
            this.comparator = comparator;
        }
        public long invoke(long a, long b) { 
            return (comparator.invoke(a, b) <= 0) ? a : b; 
        }
    }

    /**
     * A comparator for ints relying on natural ordering
     */
    public static final { int, int => int } naturalIntComparator = { int a, int b => (a < b) ? -1 : ((a > b) ? 1 : 0) };

    /**
     * A reducer returning the maximum of two int elements, using
     * natural comparator
     */
    public static final { int, int => int } naturalIntMaxReducer = { int a, int b => a >= b? a : b };

    /**
     * A reducer returning the minimum of two int elements,
     * using natural comparator
     */
    public static final { int, int => int } naturalIntMinReducer = { int a, int b => a <= b? a : b };

    /**
     * A reducer returning the maximum of two int elements,
     * using the given comparator
     */
    public static final class IntMaxReducer implements { int, int => int } {
        final { int, int => int } comparator;
        /**
         * Creates a IntMaxReducer using the given comparator
         */
        public IntMaxReducer({ int, int => int } comparator) {
            this.comparator = comparator;
        }
        public int invoke(int a, int b) { 
            return (comparator.invoke(a, b) >= 0)? a : b; 
        }
    }

    /**
     * A reducer returning the minimum of two int elements,
     * using the given comparator
     */
    public static final class IntMinReducer implements { int, int => int } {
        final { int, int => int } comparator;
        /**
         * Creates a IntMinReducer using the given comparator
         */
        public IntMinReducer({ int, int => int } comparator) {
            this.comparator = comparator;
        }
        public int invoke(int a, int b) {
            return (comparator.invoke(a, b) <= 0)? a : b;
        }
    }

    /**
     * A composite mapper that applies a second mapper to the results
     * of applying the first one
     */
    public static final class CompoundMapper<T,U,V> implements { T => V } {
        final { T => U } first;
        final { U => V } second;
        public CompoundMapper
            ({ T => U } first,
             { U => V } second) {
            this.first = first;
            this.second = second;
        }
        /** Returns <tt>second.invoke(first.invoke(t))</tt> */
        public V invoke(T t) { return second.invoke(first.invoke(t)); }
    }
        

}
